;
;*******************************************************************
; 	BMON1048.ASM                  Date 22.12.2010
;	Battery Monitor with PIC18F2520
;       by Anja
;       An_ja  @  http://www.mikrocontroller.net
;*******************************************************************
; History:
;
; BMON0945 of 27.11.2009
; - reads ADC cyclically
; - switches off load when battery voltage is below 1.15V/cell
; - activates load again when charging voltage is above 1.25V/cell
; - displays measured voltage on LCD
; - watchdog wakeup all 16 ms does multiplexing of backplane and digits
;
; BMON1035 of 09.10.2010
; - evalueates Key
; - on keypressed > 5 secs switch off/on Load
; 
; BMON1048 of 22.12.2010
; - current measurement added
; - PORT RA0 = current (was LoadOn)
; - PORT RB6 = LoadOn
;*******************************************************************
;
; PIC18F2520 is the target processor, values decimal, eXpand macros
;
	LIST	p=18F2520,r=dec,x=on
;
        include "P18F2520.INC"
;
; for Debugger MCLRE hast to be ON, Write protection OFF, WDT OFF
; BOREN=NOSLP generates Resets (see Errata Sheet)
;
	CONFIG	OSC=INTIO67,FCMEN=OFF,IESO=OFF,BOREN=OFF,BORV=3
    ifdef __DEBUG
        CONFIG  WDT=OFF,WDTPS=4,MCLRE=OFF,LPT1OSC=ON,PBADEN=OFF,DEBUG=ON
    else
        CONFIG  WDT=ON,WDTPS=4,MCLRE=OFF,LPT1OSC=ON,PBADEN=OFF,DEBUG=OFF
    endif
	CONFIG	CCP2MX=PORTC,STVREN=ON,LVP=OFF,XINST=OFF
	CONFIG	CP0=OFF,CP1=OFF,CP2=OFF,CP3=OFF,CPB=OFF,CPD=OFF
	CONFIG	WRT0=OFF,WRT1=OFF,WRT2=OFF,WRT3=OFF,WRTB=OFF,WRTC=OFF,WRTD=OFF
	CONFIG	EBTR0=OFF,EBTR1=OFF,EBTR2=OFF,EBTR3=OFF,EBTRB=OFF
;
VERSION	equ		0x1048	; ID KW 48/2010
        __idlocs       	_IDLOC0, low VERSION	; ID KW 48/2010
	__idlocs	_IDLOC1, high VERSION
;
#define	_reset		0x000   ; Reset Vector
;
;*******************************************************************
; Macros
;*******************************************************************
;
#define SKPGE	BTFSS	STATUS,C
#define	SKPLT	BTFSC	STATUS,C
#define SKPZ	BTFSS	STATUS,Z
#define SKPNZ	BTFSC	STATUS,Z
;
	include "BMON1048.H"
;
;*******************************************************************
; Programmstart + Initialisierung
;*******************************************************************
;
	org	_reset
;
LOCRESET
	NOP			;
	NOP			;
	goto	INIT		;
INTERRUPT:
	goto	LOCRESET	;
;
;*******************************************************************
; Subroutinen (Page 0!!!)
;*******************************************************************
;
	org	0x0020		;

LCD1A
	addwf	PCL,F		;
	retlw	0x40		; 0: 1F
	retlw	0x00		; 1:
	retlw	0x80		; 2; 1G
	retlw	0x80		; 3; 1G
	retlw	0xC0		; 4; 1F,1G
	retlw	0xC0		; 5; 1F,1G
	retlw	0xC0		; 6; 1F,1G
	retlw	0x00		; 7;
	retlw	0xC0		; 8; 1F,1G
	retlw	0xC0		; 9: 1F,1G

LCD1B
	addwf	PCL,F		;
	retlw	0x1C		; 0: 1C,1D,1E
	retlw	0x04		; 1: 1C
	retlw	0x18		; 2: 1D,1E
	retlw	0x0C		; 3: 1C,1D
	retlw	0x04		; 4: 1C
	retlw	0x0C		; 5: 1C,1D
	retlw	0x1C		; 6: 1C,1D,1E
	retlw	0x04		; 7: 1C
	retlw	0x1C		; 8: 1C,1D,1E
	retlw	0x0C		; 9: 1C,1D

LCD1C
	addwf	PCL,F		;
	retlw	0x03		; 0: 1A,1B
	retlw	0x02		; 1: 1B
	retlw	0x03		; 2: 1A,1B
	retlw	0x03		; 3: 1A,1B
	retlw	0x02		; 4: 1B
	retlw	0x01		; 5: 1A
	retlw	0x01		; 6: 1A
	retlw	0x03		; 7: 1A,1B
	retlw	0x03		; 8: 1A,1B
	retlw	0x03		; 9: 1A,1B
LCD2B
	addwf	PCL,F		; 
	retlw	0x03		; 0: 2E,DP2
	retlw	0x02		; 1: DP2
	retlw	0x03		; 2: 2E,DP2
	retlw	0x02		; 3: DP2
	retlw	0x02		; 4: DP2
	retlw	0x02		; 5: DP2
	retlw	0x03		; 6: 2E,DP2
	retlw	0x02		; 7: DP2
	retlw	0x03		; 8: 2E,DP2
	retlw	0x02		; 9: DP2
LCD2C
	addwf	PCL,F		;
	retlw	0xF8		; 0: 2F,2A,2B,2C,2D
	retlw	0x60		; 1: 2B,2C
	retlw	0xB4		; 2: 2G,2A,2B,   2D
	retlw	0xF4		; 3: 2G,2A,2B,2C,2D
	retlw	0x6C		; 4: 2F,2G,2B,2C
	retlw	0xDC		; 5: 2F,2G,2A,2C,2D
	retlw	0xDC		; 6: 2F,2G,2A,2C,2D
	retlw	0x70		; 7: 2A,2B,2C
	retlw	0xFC		; 8: 2F,2G, 2A,2B,2C,2D
	retlw	0xFC		; 9: 2F,2G, 2A,2B,2C,2D
;
;
INIT:
;
; Ports
;
	movlw	0xBF		; "on-Pin" = off
	movwf	LATB		;
	movlw	0x3F		;
	movwf	LATC		;
	movlw	0xC0		;
	movwf	LATA		;
;
	movlw	0x01		;
        movwf	TRISA		;
	movlw   0x00
	movwf	TRISB		;
	movwf	TRISC		;
;
; Zero Variables
;
	movlw	0x00		;
	movwf	OldState	;
	movwf	Volt		;
	movwf	Volt2		;
        movwf   KeyDeb          ;
        movwf   VoltCou         ;
	movwf	VoltOld0	;
	movwf	VoltOld1	;
	movwf	VoltOld2	;
	movwf	VoltOld3	;
	movwf	VoltOld4	;
	movwf	VoltOld5	;
	movwf	CurrCou		;
	movwf	CurrH		;
	movwf	CurrL		;
	movwf	Curr		;
	movwf	VoltH		;
	movwf	VoltL		;
;
; non-Zero variables
;
        movlw   0x03            ;
        movwf   Flags           ; switch load on
;
	movlw   VoltOld0	; VoltIdx := &VoltOld0
	movwf	VoltIdx		;
;

	movlw	0		; BARG(B0,B1) := EEPROM
	movwf	EEADR		;
	clrf	EECON1		;
	bsf	EECON1,RD	;
	movf	EEDATA,W	;
	movwf	BARGB0		;
	movwf	VoltCorrH	;

	incf	EEADR,F		; 
	bsf	EECON1,RD	;
	movf	EEDATA,W	;
	movwf	BARGB1		;
	movwf	VoltCorrL	;

	incf	EEADR,F		;
	bsf	EECON1,RD	;
	movf	EEDATA,W	;
	movwf	CurrCorrH	;

	incf	EEADR,F		;
	bsf	EECON1,RD	;
	movf	EEDATA,W	;
	movwf	CurrCorrL	;
SECLP:	
	movlw	60		;
	movwf	HZ60		;
LOOP:
	movlw	0xE0		;
	xorwf	LATA,F		;
	movlw	0x3F		;
	xorwf	LATB,F		;
	movlw	0xFF		;
	xorwf	LATC,F		;
   ifdef __DEBUG
        Delay 16 ms		;
   else
	SLEEP			;
	NOP			;
   endif
;
; check if output is off -> blinking display
;
	movf	HZ60,W		;
	xorlw	31		; 30+1 (decrement is later)
	bnz	LOOP2		;
	btfsc	LCD0,1		; if LCD0 == 2
	btg	DP1		;   DP1 := off
        btfsc   KeyDeb,0	; if KeyPressed
        btg     DP2             ;   DP2 := off
        btfsc	LoadON		;
	bra	LOOP2		;
	movlw	0x3F		;
	andwf	LATA,F		;
        movlw   0xC0		;
	andwf	LATB,F		;
	clrf	LATC		;
LOOP2
	decfsz	HZ60,F		;
	goto	LOOP		;
;
; read ADC
;
	movlw	0x1D		; RA4 + RA3 + RA2 + RA0 = input (Key + VRef + Sig + Curr)
	movwf	TRISA		;
	bsf	ExtRef		; Switch ext. reference + Mux on
	movlw	0x1B		; VREF = RA3, RA0-RA3 = analog
	movwf	ADCON1		;
	movlw	(2<<CHS0)+(1<<ADON)	; Channel 2, ADC ON
	movwf	ADCON0		;
	movlw	0x20		; select aquisition time (> 4us)
	movwf	ADCON2		;
	bsf	ADCON0,GO	;
;
; debounce Key
;
        bcf	_c              ; CY = 0
        btfss	Key		; if key pressed
	bsf	_c              ;   CY = 1
        rlcf    KeyDeb,F        ; shift to keydebounce
; 
; continue reading ADC
;
RDADC1
	btfsc	ADCON0,DONE	;
	bra	RDADC1		;
;
	movf	ADRESH,W	; store result for correction calculation
	movwf	TEMPB0		;
	movwf	VoltH		;
	movf	ADRES,W		;
	movwf	TEMPB1		;
	movwf	VoltL		;
;
; read current channel
;
	movlw	(0<<CHS0)+(1<<ADON)	; Channel 0, ADC ON
	movwf	ADCON0		;
	bsf	ADCON0,GO	;
;
	call	MulU16S16V	; calculation for voltage cannel
RDADC1a
	btfsc	ADCON0,DONE	;
	bra	RDADC1a		;
;
; store current value
;
	movf	ADRESH,W	; store result for correction calculation
	movwf	CurrH		;
	movwf	Curr		;
	movf	ADRES,W		;
	movwf	CurrL		;
;
	bcf	ADCON0,ADON	; switch ADC off
	bcf	ExtRef		; switch ext Ref + Mux off
	movlw	0x01		;
	movwf	TRISA		; A0 = current in
;
	movlw	0x80		; prepare rounding
	addwf	AARGB1,F	;
	movlw	0x00		;
	addwfc	AARGB0,F	;
;
	movf	VoltL,W		;
	addwf	AARGB1,F	;
	movff	AARGB0,AARGB1	;
	movf	VoltH,W		;
	addwfc	AARGB0,F	;
;
	btfsc	AARGB1,7	;
	bra	RDADC2
	btfsc	STATUS,C	;	
	setf	AARGB0		;
RDADC2
	movf	AARGB0,W	;
	movwf	Volt		;
;
	movlw	VOn		; 
	subwf	Volt,W		; if Volt >= VOn then _CY := 1
	rlcf	OldState,F	;   OldState += OldState + _CY
;	
	movlw	VOff		;
	subwf	Volt,W		; if Volt < VOff then _CY := 0
	rlcf	OldState,F	;   OldState += Oldstate + _CY
;
	movf	OldState,W	; if VoltOK
	btfss	VoltOK		;   W := OldState
	comf	OldState,W	; else W := ~OldState
	andlw	0x3F		; W &= 0x1F
	SKPNZ			; if W == 0
	btg	VoltOK		;   VoltOK := !VoltOK
;
; Check dU/dt < -0,5V/min every 10 seconds if Volt < VOn
;
	incf	VoltCou,F	; VoltCou ++
	movlw	VOLT_SEC	;
	subwf	VoltCou,W	; if VoltCou >= 10 then _CY := 1
	BNC	DUDT9		; if CY == 1 then do
        clrf	VoltCou		;   VoltCou := 0
;
	movf	VoltIdx,W	;
	movwf	FSR0L		;   FSR0 := VoltIdx
;
	btfsc	OldState,1	;   if Volt >= VOn then exit
	bra	DUDT8		;      (with storing current value)
;
	movf	Volt2,W		;
	subwf	INDF0,W		;
	BNC	DUDT8		;
	addlw	-VDIFF		;
	BNC	DUDT8		;
;
	movf	Volt,W		;
	subwf	INDF0,W		;   if VoltOld-Volt >= VDIFF
	BNC	DUDT8		;      (only if positive difference)
	addlw	-VDIFF		;
	BTFSC	_c		;
	BCF	VoltOK		;      VoltOK := 0
;
DUDT8
;
; store voltage value in ringbuffer
;
	movf	Volt,W		;  @VoltIdx := Volt
	movwf	INDF0		;
	incf	VoltIdx,F	;  VoltIdx ++
	movlw	VoltOld5+1	;  
	subwf	VoltIdx,W	;  if VoltIdx >= VoltOld5+1 then _CY := 1
	movlw	VoltOld0	;
	BTFSC	_c		;  if _CY == 1 then 
	movwf	VoltIdx		;    VoltIdx := VoltOld0
DUDT9
	movf	Volt,W		;
	movwf	Volt2		;
;
; evaluate Key
; first: debounce
;
	movf	KeyDeb,W	;
	andlw	0x07		;
	xorlw	0x05		;
	SKPNZ			;
	bsf	KeyDeb,1	;
	xorlw	0x07		;
	SKPNZ			;
	bcf	KeyDeb,1	;
;
; toggle output after 5-6 secs keypressed
;
	movf	KeyDeb,W	;
	xorlw	0x3F		;
        SKPNZ			;
	btg	KeyOn		;
;
; set output according to flags
;
	movf	Flags,W		;
	andlw	0x03		;
	xorlw	0x03		; if Flags == 3
	SKPNZ  			;
	bsf	LoadON		; then LoadON = 1
	SKPZ			;
	bcf	LoadON          ; else LoadON = 0
;
; Check if current display is necessary
;
	bcf	CurrOK		;
	movlw	CURR_MIN	;
	subwf	Curr,W		; if Curr >= CURR_MIN then _CY := 1
	BNC	DspC1		; if CY == 1 then do
        bsf	CurrOK		;   CurrOK := 1
DspC1	
;
	incf	CurrCou,F	;
	btfsc	CurrCou,1	; Display current only every two of four seconds
	bcf	CurrOK		;

	btfss	CurrOK		;
	bra	DspC3		;
;
; Correction of Current Value	;
;
	movf	CurrH,W		;
	movwf	TEMPB0		;
	movf	CurrL,W		;
	movwf	TEMPB1		;

	call	MulU16S16C	; calculation for current cannel
;
	movlw	0x80		; prepare rounding
	addwf	AARGB1,F	;
	movlw	0x00		;
	addwfc	AARGB0,F	;
;
	movf	CurrL,W		;
	addwf	AARGB1,F	;
	movff	AARGB0,AARGB1	;
	movf	CurrH,W		;
	addwfc	AARGB0,F	;
;
	btfsc	AARGB1,7	;
	bra	DspC2
	btfsc	STATUS,C	;	
	setf	AARGB0		;
DspC2
	movf	AARGB0,W	;
	movwf	Curr		;	
;
DspC3
	movlw	LCD0		;
	movwf	FSR0L		;
	movf	Volt,W		; Voltage
	btfsc	CurrOK		;
	movf	Curr,W		; Current
	call	Bin2BCD_3	;
	movlw	LCD0		;
	movwf	FSR0L		;
BCD2SEG
	clrf	SEGA		;
	clrf	SEGB		;
	clrf	SEGC		;
	movf	LCD0,F		;
	bz	BCD2SEG1	;
	bsf	SEGB,5		;
BCD2SEG1
	rlncf	LCD1,W		;
;	clrf	PCLATH		;
	call	LCD1A		;
	iorwf	SEGA		;

	rlncf	LCD1,W		;
	call	LCD1B		;
	iorwf	SEGB		;

	rlncf	LCD1,W		;
	call	LCD1C		;
	iorwf	SEGC		;

	rlncf	LCD2,W		;
	call	LCD2B		;
	iorwf	SEGB		;
	btfsc	CurrOK		;
	bcf	SEGB,1		; Current Display without decimal point


	rlncf	LCD2,W		;
	call	LCD2C		;
	iorwf	SEGC		;
;
	movf	LATA,W		;
	andlw	0x1F		;
	iorwf	SEGA,W		;
	movwf	LATA		;
;
	movf	LATB,W		;
	andlw	0xC0		;
	iorwf	SEGB,W		;
	movwf	LATB		;
;
	movff	SEGC,LATC	;
	goto	SECLP		;


Bin2BCD_3
;
; Calculate BCD-Value (3 Digits) out of Binary Byte
;
	movwf	AARGB0		;
	setf	INDF0		; Result := -1
	movlw	100		;
Bin2BCD1
	incf	INDF0,F		; Increment Result	
	subwf	AARGB0,F	;
	bc	Bin2BCD1	;
Bin2BCD2
	addwf	AARGB0,F	; correction of last step
	incf	FSR0L,F		; next digit
	setf	INDF0		; Result := -1
	movlw	10		;
Bin2BCD3
	incf	INDF0,F		;
	subwf	AARGB0,F	;
	bc	Bin2BCD3	;
Bin2BCD4
	addwf	AARGB0,F	; correction of last step
	incf	FSR0L,F		; next digit
	movf	AARGB0,W	;
	movwf	INDF0		;
	return

MulU16S16C
;
; Current correction
;
	movf	CurrCorrH,W	;
	movwf	BARGB0		;
	movf	CurrCorrL,W	;
	movwf	BARGB1		;
	bra	MulU16S16	;
;
MulU16S16V
;
; Voltage correction
;
	movf	VoltCorrH,W	;
	movwf	BARGB0		;
	movf	VoltCorrL,W	;
	movwf	BARGB1		;
;
MulU16S16
	movf	TEMPB1,W	; low-bytes
	mulwf	BARGB1		;
	movff	PRODH,AARGB2	;
	movff	PRODL,AARGB3	;
;
	movf	TEMPB0,W	; high-bytes
	mulwf	BARGB0		;
	movff	PRODH,AARGB0	;
	movff	PRODL,AARGB1	;
;
	movf	TEMPB1,W
	mulwf	BARGB0		;
	movf	PRODL,W		;
	addwf	AARGB2,F	; Add cross
	movf	PRODH,W 	; products
	addwfc	AARGB1,F	;
	clrf	WREG		;
	addwfc	AARGB0,F 	;
;
	movf	TEMPB0,W	;
	mulwf	BARGB1		;
	movf	PRODL,W		;
	addwf	AARGB2,F	; Add cross
	movf	PRODH,W		; products
	addwfc	AARGB1,F	;
	clrf	WREG		;
	addwfc	AARGB0,F 	;
;
	btfss	BARGB0,7 	; BARG neg?
	bra	MulU16S16_E 	; no, check ARG1
	movf	TEMPB1,W	;
	subwf	AARGB1		;
	movf	TEMPB0,W	;
	subwfb	AARGB0		;
;
MulU16S16_E
;
	return
;
;*******************************************************************
; Delay8
;
; performance: 
;   4 + 4*W cycles (including call + ret)
;*******************************************************************
;	
Delay8:
	movwf	TEMP		;
Delay81:
	nop			;
	decfsz	TEMP,F		;
	bra	Delay81		;
	return			;
;
Delay16:			;
	clrf	TEMP		;
Delay161			;
	decfsz	TEMP,F		;inner loop: Temp
	bra	Delay161	;
	clrwdt			;
;
	addlw	-1		;outer loop: ACCU
				;
	bnz	Delay161	;
	return			;
;

	org  0xF00000		; EEPROM defaults
;
; Default values for nominal resistors
;
	de	0x00,0x00	; Voltage correction +0.00%
	de	0xF3,0x1B	; Current correction -5.04%

; Sample #2,#3
;    de  0x01,0xE3   ; Correction +0.74%
;    de  0xEE,0x8B   ; Current correction -6.8%

; Sample #1
;	de	0xFE,0x24	; Correction -0.73%
;        de      0xF1,0x12       ; Current Correction -6%

	end
